# -*- coding: utf-8 -*-

from __future__ import unicode_literals

COLORS = [BLACK, WHITE] = range(2)

PIECE_TYPES_WITH_NONE = [NONE,
           PAWN,      LANCE,      KNIGHT,      SILVER,
           GOLD,
         BISHOP,       ROOK,
           KING,
      PROM_PAWN, PROM_LANCE, PROM_KNIGHT, PROM_SILVER,
    PROM_BISHOP,  PROM_ROOK,
] = range(15)
PIECE_TYPES = [
           PAWN,      LANCE,      KNIGHT,      SILVER,
           GOLD,
         BISHOP,       ROOK,
           KING,
      PROM_PAWN, PROM_LANCE, PROM_KNIGHT, PROM_SILVER,
    PROM_BISHOP,  PROM_ROOK,
]



# -*- coding: utf-8 -*-
#
# This file is part of the python-shogi library.
# Copyright (C) 2015- Tasuku SUENAGA <tasuku-s-github@titech.ac>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals

import re
import time
import shogi
import socket
import threading
import collections

DEFAULT_PORT = 4081
PING_SLEEP_DURATION = 1
PING_DURATION = 60
SOCKET_RECV_SIZE = 4096
BLOCK_RECV_SLEEP_DURATION = 0.1

COLOR_SYMBOLS = ['+', '-']
PIECE_SYMBOLS = ['* ', 'FU', 'KY', 'KE', 'GI', 'KI', 'KA', 'HI', 'OU',
                       'TO', 'NY', 'NK', 'NG',       'UM', 'RY']
SQUARE_NAMES = [
    '91', '81', '71', '61', '51', '41', '31', '21', '11',
    '92', '82', '72', '62', '52', '42', '32', '22', '12',
    '93', '83', '73', '63', '53', '43', '33', '23', '13',
    '94', '84', '74', '64', '54', '44', '34', '24', '14',
    '95', '85', '75', '65', '55', '45', '35', '25', '15',
    '96', '86', '76', '66', '56', '46', '36', '26', '16',
    '97', '87', '77', '67', '57', '47', '37', '27', '17',
    '98', '88', '78', '68', '58', '48', '38', '28', '18',
    '99', '89', '79', '69', '59', '49', '39', '29', '19',
]
SERVER_MESSAGE_SYMBOLS = [
    # '#' prefixed
    'WIN', 'LOSE', 'DRAW', 'SENNICHITE', 'OUTE_SENNICHITE',
    'ILLEGAL_MOVE', 'TIME_UP', 'RESIGN', 'JISHOGI', 'CHUDAN',
    'MAX_MOVES', 'CENSORED',
    # '%' prefixed
    'TORYO', 'KACHI'
]
SERVER_MESSAGES = [
    WIN, LOSE, DRAW, SENNICHITE, OUTE_SENNICHITE,
    ILLEGAL_MOVE, TIME_UP, REGISN, JISHOGI, CHUDAN,
    MAX_MOVES, CENSORED,
    TORYO, KACHI
] = range(0, len(SERVER_MESSAGE_SYMBOLS))

class Parser:
    @staticmethod
    def parse_file(path):
        with open(path) as f:
            return Parser.parse_str(f.read())

    @staticmethod
    def parse_str(csa_str):
        line_no = 1

        sfen = None
        board = None
        position_lines = []
        names = [None, None]
        current_turn_str = None
        moves = []
        lose_color = None
        for line in csa_str.split('\n'):
            if line == '':
                pass
            elif line[0] == "'":
                pass
            elif line[0] == 'V':
                # Currently just ignoring version
                pass
            elif line[0] == 'N' and line[1] in COLOR_SYMBOLS:
                names[COLOR_SYMBOLS.index(line[1])] = line[2:]
            elif line[0] == '$':
                # Currently just ignoring information
                pass
            elif line[0] == 'P':
                position_lines.append(line)
            elif line[0] in COLOR_SYMBOLS:
                if len(line) == 1:
                    current_turn_str = line[0]
                else:
                    if not board:
                        raise ValueError('Board infomation is not defined before a move')
                    (color, move) = Parser.parse_move_str(line, board)
                    moves.append(move)
                    board.push(shogi.Move.from_usi(move))
            elif line[0] == 'T':
                # Currently just ignoring consumed time
                pass
            elif line[0] == '%':
                # End of the game
                if not board:
                    raise ValueError('Board infomation is not defined before a special move')
                if line in [
                            '%TORYO', '%TIME_UP', '%ILLEGAL_MOVE'
                        ]:
                    lose_color = board.turn
                elif line == '%+ILLEGAL_ACTION':
                    lose_color = shogi.BLACK
                elif line == '%-ILLEGAL_ACTION':
                    lose_color = shogi.WHITE

                # TODO: Support %MATTA etc.
                break
            elif line == '/':
                raise ValueError('Dont support multiple matches in str')
            else:
                raise ValueError('Invalid line {0}: {1}'.format(line_no, line))
            if board is None and current_turn_str:
                position = Parser.parse_position(position_lines)
                sfen = Exporter.sfen(
                    position['pieces'],
                    position['pieces_in_hand'],
                    current_turn_str,
                    1)
                board = shogi.Board(sfen)
            line_no += 1

        if lose_color == shogi.BLACK:
            win = 'w'
        elif lose_color == shogi.WHITE:
            win = 'b'
        else:
            win = '-'

        summary = {
            'names': names,
            'sfen': sfen,
            'moves': moves,
            'win': win
        }
        # NOTE: for future support of multiple matches
        return [summary]

    @staticmethod
    def parse_move_str(move_str, board):
        color = COLOR_SYMBOLS.index(move_str[0])
        from_str = move_str[1:3]
        to_str = move_str[3:5]
        piece_str = move_str[5:7]

        if from_str == '00':
            from_square = None
        else:
            from_square = SQUARE_NAMES.index(from_str)

        to_square = SQUARE_NAMES.index(to_str)
        piece_type = PIECE_SYMBOLS.index(piece_str)

        if from_square is None:
            return (color, '{0}*{1}'.format(shogi.PIECE_SYMBOLS[piece_type].upper(),
                shogi.SQUARE_NAMES[to_square]))
        else:
            from_piece_type = board.pieces[from_square]
            promotion = (from_piece_type != piece_type)
            return (color, shogi.SQUARE_NAMES[from_square] + shogi.SQUARE_NAMES[to_square] + ('+' if promotion else ''))

    @staticmethod
    def parse_position(position_block_lines):
        # ex.) P1-KY-KE-GI-KI-OU-KI-GI-KE-KY
        # ex.) PI82HI22KA
        position = {
            'pieces': [None for x in range(81)],
            'pieces_in_hand': [
                collections.Counter(),
                collections.Counter(),
            ]
        }
        for line in position_block_lines:
            if line[0] != 'P':
                if line[0] in COLOR_SYMBOLS:
                    color = COLOR_SYMBOLS.index(line[0])
                    if len(line) == 1:
                        # duplicated data
                        position['current_turn'] = color
                    else:
                        # move
                        raise NotImplementedError('TODO: parse moves')
                else:
                    raise ValueError('Invalid position line: {0}'.format(line))
            elif line[1] in COLOR_SYMBOLS:
                index = 2
                color = COLOR_SYMBOLS.index(line[1])
                while index < len(line):
                    file_index = int(line[index:index + 1])
                    index += 1
                    rank_index = int(line[index:index + 1])
                    index += 1
                    piece_type = PIECE_SYMBOLS.index(line[index:index + 2])
                    index += 2
                    if rank_index == 0 and file_index == 0:
                        # piece in hand
                        position['pieces_in_hand'][color][piece_type] += 1
                    else:
                        position['pieces'][
                            (rank_index - 1) * 9 + (9 - file_index)
                        ] = (piece_type, color)
            elif line[1] in ['1', '2', '3', '4', '5', '6', '7', '8', '9']:
                rank_index = int(line[1:2]) - 1
                file_index = 0
                for index in range(2, 29, 3):
                    piece_str = line[index:index + 3]
                    piece_type = PIECE_SYMBOLS.index(piece_str[1:3])
                    if piece_type:
                        color = COLOR_SYMBOLS.index(piece_str[0])
                        piece = (piece_type, color)
                    else:
                        piece = None

                    position['pieces'][
                        rank_index * 9 + file_index
                    ] = piece

                    file_index += 1
            elif line[1] == 'I': # PI
                position['pieces'] = [
                        (2, 1), (3, 1), (4, 1), (5, 1), (8, 1), (5, 1), (4, 1), (3, 1), (2, 1),
                          None, (7, 1),   None,   None,   None,   None,   None, (6, 1),   None,
                        (1, 1), (1, 1), (1, 1), (1, 1), (1, 1), (1, 1), (1, 1), (1, 1), (1, 1),
                          None,   None,   None,   None,   None,   None,   None,   None,   None,
                          None,   None,   None,   None,   None,   None,   None,   None,   None,
                          None,   None,   None,   None,   None,   None,   None,   None,   None,
                        (1, 0), (1, 0), (1, 0), (1, 0), (1, 0), (1, 0), (1, 0), (1, 0), (1, 0),
                          None, (6, 0),   None,   None,   None,   None,   None, (7, 0),   None,
                        (2, 0), (3, 0), (4, 0), (5, 0), (8, 0), (5, 0), (4, 0), (3, 0), (2, 0),
                    ]
                index = 2
                while index < len(line):
                    file_index = int(line[index:index + 1])
                    index += 1
                    rank_index = int(line[index:index + 1])
                    index += 1
                    piece_type = PIECE_SYMBOLS.index(line[index:index + 2])
                    index += 2
                    if rank_index == 0 and file_index == 0:
                        # piece in hand
                        raise NotImplementedError('TODO: Not implemented komaochi in komadai')
                    piece_index = (rank_index - 1) * 9 + (9 - file_index)
                    if position['pieces'][piece_index] is None or position['pieces'][piece_index][0] != piece_type:
                        raise ValueError('Invalid piece removing on intializing a board'.format(line))
                    position['pieces'][piece_index] = None
            else:
                raise ValueError('Invalid rank/piece in hand: {0}'.format(line))
        position['pieces_in_hand'] = [
            dict(position['pieces_in_hand'][0]),
            dict(position['pieces_in_hand'][1]),
        ]
        return position

class Exporter:
    @staticmethod
    def sfen(pieces, pieces_in_hand, current_turn_char, move_count):
        sfen = []
        empty = 0

        # Position part.
        for square in shogi.SQUARES:
            piece_tuple = pieces[square]
            if piece_tuple is None:
                empty += 1
            else:
                (piece_type, color) = piece_tuple
                piece = shogi.Piece(piece_type, color)

                if empty:
                    sfen.append(str(empty))
                    empty = 0
                sfen.append(piece.symbol())

            if shogi.BB_SQUARES[square] & shogi.BB_FILE_1:
                if empty:
                    sfen.append(str(empty))
                    empty = 0

                if square != shogi.I1:
                    sfen.append('/')

        sfen.append(' ')

        # Side to move.
        current_turn = COLOR_SYMBOLS.index(current_turn_char)
        if current_turn == shogi.WHITE:
            sfen.append('w')
        else:
            sfen.append('b')

        sfen.append(' ')

        # Pieces in hand
        pih_len = 0
        for color in shogi.COLORS:
            p = pieces_in_hand[color]
            pih_len += len(p)
            for piece_type in p.keys():
                if p[piece_type] > 1:
                    sfen.append(str(p[piece_type]))
                elif p[piece_type] >= 1:
                    piece = shogi.Piece(piece_type, color)
                    sfen.append(piece.symbol())
        if pih_len == 0:
            sfen.append('-')

        sfen.append(' ')

        # Move count
        sfen.append(str(move_count))

        sfen_str = ''.join(sfen)

        return sfen_str

class TCPProtocol:
    def __init__(self, host=None, port=0):
        if host:
            self.open(host, port)

    def open(self, host, port=0):
        if not port:
            port = DEFAULT_PORT
        self.host = host
        self.port = port

        self.recv_buf = ''

        self.connect(host, port)

        # Heartbeats
        self.heartbeat_thread = CSAHeartbeat(self, PING_SLEEP_DURATION, PING_DURATION)
        self.heartbeat_thread.start()

    def connect(self, host, port):
        for res in socket.getaddrinfo(host, port, 0, socket.SOCK_STREAM):
            af, socktype, proto, canonname, sa = res
            try:
                self.socket = socket.socket(af, socktype, proto)
                self.socket.connect(sa)
            except socket.error as msg:
                self.msg = msg
                if self.socket:
                    self.socket.close()
                self.socket = None
                continue
            break
        if not self.socket:
            raise socket.error(self.msg)

    def command(self, command):
        self.write(command + '\n')
        line = self.read_line()
        return line

    def write(self, buf):
        self.socket.sendall(buf.encode('utf-8'))

    def read(self):
        buf = self.socket.recv(SOCKET_RECV_SIZE).decode('utf-8')
        self.recv_buf += buf
        return len(buf)

    def read_line(self, block=True):
        line = self.read_until('\n', block)
        return line

    def read_until(self, target, block=True):
        while 1:
            if target in self.recv_buf:
                (result, self.recv_buf) = self.recv_buf.split(target, 1)
                return result
            else:
                if self.read() == 0:
                    if block:
                        time.sleep(BLOCK_RECV_SLEEP_DURATION)
                    else:
                        return None

    def ping(self):
        line = self.command('')
        if line != '':
            raise ValueError('Ping return must be empty')

    login_username_re = re.compile(r'\A[-_0-9A-Za-z]+\Z')
    login_response_re = re.compile(r'\ALOGIN:([-_0-9A-Za-z]+)( OK)?\Z')

    def login(self, username, password):
        if not self.login_username_re.match(username):
            raise ValueError('Invalid username.')
        if ' ' in password:
            raise ValueError('Invalid password.')

        line = self.command('LOGIN {0} {1}'.format(username, password))
        line_match = self.login_response_re.match(line)
        if line_match:
            if line_match.group(2) == ' OK':
                if line_match.group(1) == username:
                    return True
            elif line_match.group(1) == 'incorrect':
                raise ValueError('Login failed. Check username and password.')
        raise ValueError('Login response was invalid.')

    def login_ex(self, username, password):
        if not self.login_username_re.match(username):
            raise ValueError('Invalid username.')
        if ' ' in password:
            raise ValueError('Invalid password.')

        line = self.command('LOGIN {0} {1} x1'.format(username, password))
        line_match = self.login_response_re.match(line)
        if line_match:
            if line_match.group(2) == ' OK':
                if line_match.group(1) == username:
                    return True
            elif line_match.group(1) == 'incorrect':
                raise ValueError('Login failed. Check username and password.')
        raise ValueError('Login response was invalid.')
        
    def logout(self):
        line = self.command('LOGOUT')
        if line == 'LOGOUT:completed':
            raise ValueError('Logout failed')

    def wait_match(self, block=True):
        while True:
            game_summary_str = self.read_game_summary(block)
            if game_summary_str is not None:
                return self.parse_game_summary(game_summary_str)
            else:
                return None

    def wait_server_message(self, board, block=True):
        while True:
            line = self.read_line(block)
            if line is None:
                return None
            return self.parse_server_message(line, board)

    def parse_server_message(self, line, board):
        if line[0] in COLOR_SYMBOLS:
            move_strs = line.split(',')
            move_str = move_strs[0]
            time_str = move_strs[1] if len(move_strs) > 1 else None
            (color, usi) = Parser.parse_move_str(move_str, board)
            return (color, usi, self.parse_consumed_time_str(time_str), None)
        elif line[0] in ['#', '%']:
            message = SERVER_MESSAGE_SYMBOLS.index(line[1:])
            return (None, None, None, message)
        else:
            raise ValueError('Invalid lines')

    def parse_consumed_time_str(self, time_str):
        # This function always returns float seconds.
        # TODO: refer Time_Unit header.
        if time_str is None:
            return None
        elif time_str[0] != 'T':
            raise ValueError('Invalid consumed time format')
        return float(time_str[1:])

    def read_game_summary(self, block=True):
        return self.read_until('END Game_Summary\n', block)

    def parse_game_summary(self, game_summary_block):
        time_lines = None
        position_lines = None
        names = [None, None]
        position = None
        for line in game_summary_block.split('\n'):
            if line == 'BEGIN Game_Summary' or line == 'END Game_Summary':
                pass
            elif line == 'BEGIN Time':
                time_lines = []
            elif line == 'END Time':
                time_summary = self.parse_time(time_lines)
                time_lines = None
            elif line == 'BEGIN Position':
                position_lines = []
            elif line == 'END Position':
                position = Parser.parse_position(position_lines)
                position_lines = None
            elif time_lines is not None:
                time_lines.append(line)
            elif position_lines is not None:
                position_lines.append(line)
            elif ':' in line:
                (key, value) = line.split(':', 1)
                if key == 'Name+':
                    # sente or shiatte
                    names[shogi.BLACK] = value
                elif key == 'Name-':
                    # sente or shiatte
                    names[shogi.WHITE] = value
                elif key == 'To_Move':
                    to_move_color_str = value
                elif key == 'Your_Turn':
                    my_color = COLOR_SYMBOLS.index(value)
            elif not line:
                # empty line
                pass
            else:
                raise ValueError('Invalid game summary line: {0}'.format(line))

        sfen = Exporter.sfen(
            position['pieces'],
            position['pieces_in_hand'],
            to_move_color_str,
            1)

        summary = {
            'names': names,
            'sfen': sfen,
            'moves': [],
            'time': time_summary,
        }

        return {
            'summary': summary,
            'my_color': my_color,
        }

    def parse_time(self, time_block_lines):
        time = {}
        for line in time_block_lines:
            (key, value) = line.split(':', 1)
            time[key] = value
        return time

    def agree(self):
        # TODO: check START:<GameID>
        self.command('AGREE')

    def reject(self):
        # TODO: check REJECT:<GameID> by <rejector>
        self.command('REJECT')

    def move(self, piece_type, color, move):
        if move.from_square is None:
            from_square = '00'
        else:
            from_square = SQUARE_NAMES[move.from_square]
        command = '{0}{1}{2}{3}'.format(COLOR_SYMBOLS[color],
                from_square,
                SQUARE_NAMES[move.to_square],
                PIECE_SYMBOLS[piece_type])
        return self.command(command)

    def resign(self):
        # TODO: check RESIGN
        self.command('%TORYO')
        match_result = self.read_line()

class CSAHeartbeat(threading.Thread):
    def __init__(self, ping_target, sleep_duration, ping_duration):
        super(CSAHeartbeat, self).__init__()
        self.ping_timer = 0
        self.ping_target = ping_target
        self.sleep_duration = sleep_duration
        self.ping_duration = ping_duration

    def run(self):
        if self.ping_timer >= self.ping_duration:
            self.ping_target.ping()

        self.ping_timer += self.sleep_duration
        time.sleep(self.sleep_duration)



# -*- coding: utf-8 -*-
#
# This file is part of the python-shogi library.
# Copyright (C) 2015- Tasuku SUENAGA <tasuku-s-github@titech.ac>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

# NOTE: Don't support ki2(Kifu2) format

from __future__ import unicode_literals

import os
import re
import sys
import shogi
import codecs

# python dict pre-version 3.7 is not guaranteed to be sorted by insertion.
# The OrderedDict implementation fixes this for the older versions.
if sys.version_info[0] == 2 or (sys.version_info[0] == 3 and sys.version_info[1] < 7):
    from collections import OrderedDict as ordered_dict
else:
    ordered_dict = dict

class ParserException(Exception):
    pass

class Parser:
    MOVE_RE = re.compile(r'\A *[0-9]+\s+(中断|投了|持将棋|先日手|詰み|切れ負け|反則勝ち|反則負け|(([１２３４５６７８９])([零一二三四五六七八九])|同　)([歩香桂銀金角飛玉と杏圭全馬龍])(打|(成?)\(([0-9])([0-9])\)))\s*(\([ /:0-9]+\))?\s*\Z')

    HANDYCAP_SFENS = {
        '平手': shogi.STARTING_SFEN,
        '香落ち': 'lnsgkgsn1/1r5b1/ppppppppp/9/9/9/PPPPPPPPP/1B5R1/LNSGKGSNL w - 1',
        '右香落ち': '1nsgkgsnl/1r5b1/ppppppppp/9/9/9/PPPPPPPPP/1B5R1/LNSGKGSNL w - 1',
        '角落ち': 'lnsgkgsnl/1r7/ppppppppp/9/9/9/PPPPPPPPP/1B5R1/LNSGKGSNL w - 1',
        '飛車落ち': 'lnsgkgsnl/7b1/ppppppppp/9/9/9/PPPPPPPPP/1B5R1/LNSGKGSNL w - 1',
        '飛香落ち': 'lnsgkgsn1/7b1/ppppppppp/9/9/9/PPPPPPPPP/1B5R1/LNSGKGSNL w - 1',
        '二枚落ち': 'lnsgkgsnl/9/ppppppppp/9/9/9/PPPPPPPPP/1B5R1/LNSGKGSNL w - 1',
        '三枚落ち': 'lnsgkgsn1/9/ppppppppp/9/9/9/PPPPPPPPP/1B5R1/LNSGKGSNL w - 1',
        '四枚落ち': '1nsgkgsn1/9/ppppppppp/9/9/9/PPPPPPPPP/1B5R1/LNSGKGSNL w - 1',
        '五枚落ち': '2sgkgsn1/9/ppppppppp/9/9/9/PPPPPPPPP/1B5R1/LNSGKGSNL w - 1',
        '左五枚落ち': '1nsgkgs2/9/ppppppppp/9/9/9/PPPPPPPPP/1B5R1/LNSGKGSNL w - 1',
        '六枚落ち': '2sgkgs2/9/ppppppppp/9/9/9/PPPPPPPPP/1B5R1/LNSGKGSNL w - 1',
        '八枚落ち': '4k4/9/ppppppppp/9/9/9/PPPPPPPPP/1B5R1/LNSGKGSNL w - 1',
        'その他': None
    }

    RESULT_RE = re.compile(r'　*まで(\d+)手で((先|下|後|上)手の勝ち|千日手|持将棋|中断)')

    @staticmethod
    def parse_file(path):
        prefix, ext = os.path.splitext(path)
        for enc in ['cp932', 'utf-8-sig']:
            try:
                with codecs.open(path, 'r', enc) as f:
                    return Parser.parse_str(f.read())
            except:
                pass
        return None

    @staticmethod
    def parse_pieces_in_hand(target):
        if target == 'なし': # None in japanese
            return ordered_dict()

        result = ordered_dict()
        for item in target.split('　'):
            if len(item) == 1:
                result[shogi.PIECE_JAPANESE_SYMBOLS.index(item)] = 1
            elif len(item) == 2 or len(item) == 3:
                result[shogi.PIECE_JAPANESE_SYMBOLS.index(item[0])] = \
                    shogi.NUMBER_JAPANESE_KANJI_SYMBOLS.index(item[1:])
            elif len(item) == 0:
                pass
            else:
                raise ParserException('Invalid pieces in hand')
        return result

    @staticmethod
    def parse_board_line(line):
        board_line = line.split('|')[1].replace(' ', '')
        line_sfen = ''
        square_skip = 0
        sente = True

        for square in board_line:
            # if there is a piece in the square (no dot)
            if square != '・':
                # if there is a square skip, add to sfen
                if square_skip > 0:
                    line_sfen = ''.join((line_sfen, str(square_skip)))
                    square_skip = 0

                if square == 'v':
                    sente = False
                    continue

                # get the piece roman symbol
                piece = shogi.PIECE_SYMBOLS[
                    shogi.PIECE_JAPANESE_SYMBOLS.index(square[-1])]

                # if sente
                if sente:
                    line_sfen = ''.join((line_sfen, piece.upper()))
                else:
                    line_sfen = ''.join((line_sfen, piece.lower()))
                sente = True
            else:
                square_skip += 1

        # if last square is also empty, need to add the skip to the end
        if square_skip > 0:
            line_sfen = ''.join((line_sfen, str(square_skip)))

        return line_sfen

    @staticmethod
    def complete_custom_sfen(board, pieces_in_hand, turn):
        if turn == shogi.BLACK:
            turn_str = 'b'
        else:
            turn_str = 'w'

        # add whos turn it is
        sfen = ''.join((board, ' ', turn_str, ' '))

        # if there are pieces in the hand
        if sum(pieces_in_hand[shogi.BLACK].values()) + \
           sum(pieces_in_hand[shogi.WHITE].values()):
            for key, quantity in pieces_in_hand[shogi.BLACK].items():
                piece = shogi.PIECE_SYMBOLS[key].upper()
                if quantity > 1:
                    sfen = ''.join((sfen, str(quantity), piece))
                elif quantity == 1:
                    sfen = ''.join((sfen, piece))

            for key, quantity in pieces_in_hand[shogi.WHITE].items():
                piece = shogi.PIECE_SYMBOLS[key].lower()
                sfen = ''.join((sfen, str(quantity), piece))
        else:
            sfen = ''.join((sfen, '-'))

        # add the initial move number
        sfen = ''.join((sfen, ' 1'))

        return sfen

    @staticmethod
    def parse_move_str(line, last_to_square):
        # Normalize king/promoted kanji
        line = line.replace('王', '玉')
        line = line.replace('竜', '龍')
        line = line.replace('成銀', '全')
        line = line.replace('成桂', '圭')
        line = line.replace('成香', '杏')

        m = Parser.MOVE_RE.match(line)
        if m:
            if m.group(1) in [
                    '中断',
                    '投了',
                    '持将棋',
                    '千日手',
                    '詰み',
                    '切れ負け',
                    '反則か勝ち',
                    '反則負け'
            ]:
                return (
                    None,
                    None,
                    m.group(1)
                )

            piece_type = shogi.PIECE_JAPANESE_SYMBOLS.index(m.group(5))
            if m.group(2) == '同　':
                # same position
                to_square = last_to_square
            else:
                to_field = 9 - shogi.NUMBER_JAPANESE_NUMBER_SYMBOLS.index(m.group(3))
                to_rank = shogi.NUMBER_JAPANESE_KANJI_SYMBOLS.index(m.group(4)) - 1
                to_square = to_rank * 9 + to_field
                last_to_square = to_square

            if m.group(6) == '打':
                # piece drop
                return (
                    '{0}*{1}'.format(shogi.PIECE_SYMBOLS[piece_type].upper(), shogi.SQUARE_NAMES[to_square]),
                    last_to_square,
                    None
                )
            else:
                from_field = 9 - int(m.group(8))
                from_rank = int(m.group(9)) - 1
                from_square = from_rank * 9 + from_field

                promotion = (m.group(7) == '成')
                return (
                    shogi.SQUARE_NAMES[from_square] + shogi.SQUARE_NAMES[to_square] + ('+' if promotion else ''),
                    last_to_square,
                    None
                )
        return (None, last_to_square, None)

    @staticmethod
    def parse_str(kif_str):
        line_no = 1

        names = [None, None]
        pieces_in_hand = [ordered_dict(), ordered_dict()]
        current_turn = shogi.BLACK
        sfen = shogi.STARTING_SFEN
        moves = []
        last_to_square = None
        win = None
        custom_sfen = False
        kif_str = kif_str.replace('\r\n', '\n').replace('\r', '\n')
        for line in kif_str.split('\n'):
            if len(line) == 0 or line[0] == "*":
                pass
            elif line.count('+') == 2 and line.count('-') > 10:
                if custom_sfen:
                    custom_sfen = False
                    # remove last slash
                    sfen = sfen[:-1]
                else:
                    custom_sfen = True
                    sfen = ''
            elif custom_sfen:
                sfen = ''.join((sfen, Parser.parse_board_line(line), '/'))
            elif '：' in line:
                (key, value) = line.split('：', 1)
                value = value.rstrip('　')
                if key == '先手' or key == '下手': # sente or shitate
                    # Blacks's name
                    names[shogi.BLACK] = value
                elif key == '後手' or key == '上手': # gote or uwate
                    # White's name
                    names[shogi.WHITE] = value
                elif key == '先手の持駒' or \
                        key == '下手の持駒': # sente or shitate's pieces in hand
                    # First player's pieces in hand
                    pieces_in_hand[shogi.BLACK] = Parser.parse_pieces_in_hand(value)
                elif key == '後手の持駒' or \
                        key == '上手の持駒': # gote or uwate's pieces in hand
                    # Second player's pieces in hand
                    pieces_in_hand[shogi.WHITE] = Parser.parse_pieces_in_hand(value)
                elif key == '手合割': # teai wari
                    sfen = Parser.HANDYCAP_SFENS[value]
                    if sfen is None:
                        raise ParserException('Cannot support handycap type "other"')
            elif line == '後手番':
                # Current turn is white
                current_turn = shogi.WHITE
            else:
                (move, last_to_square, special_str) = Parser.parse_move_str(line, last_to_square)
                if move is not None:
                    moves.append(move)
                    if current_turn == shogi.BLACK:
                        current_turn = shogi.WHITE
                    else: # current_turn == shogi.WHITE
                        current_turn = shogi.BLACK
                elif special_str in ['投了', '詰み', '切れ負け', '反則負け']:
                    if current_turn == shogi.BLACK:
                        win = 'w'
                    else: # current_turn == shogi.WHITE
                        win = 'b'
                elif special_str in ['反則勝ち', '入玉勝ち']:
                    if current_turn == shogi.BLACK:
                        win = 'b'
                    else: # current_turn == shogi.WHITE
                        win = 'w'
                elif special_str in ['持将棋', '先日手']:
                    win = '-'
                else:
                    m = Parser.RESULT_RE.match(line)
                    if m:
                        win_side_str = m.group(3)
                        if win_side_str == '先' or win_side_str == '下':
                            win = 'b'
                        elif win_side_str == '後' or win_side_str == '上':
                            win = 'w'
                        else:
                            # TODO: repetition of moves with continuous check
                            win = '-'
            line_no += 1

        # if using a custom sfen
        if len(sfen.split(' ')) == 1:
            sfen = Parser.complete_custom_sfen(sfen, pieces_in_hand, current_turn)

        summary = {
            'names': names,
            'sfen': sfen,
            'moves': moves,
            'win': win
        }

        # NOTE: for the same interface with CSA parser
        return [summary]



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

# NOTE: In chess we use "file - rank" notation like 'a1',
#       in shogi we use a number for file, an alphabet for rank
#       and opposite direction of files and ranks like '9i'.
#       We use chess style notation internally, but exports it with this table.

from .Piece import *

SQUARE_NAMES = [
    '9a', '8a', '7a', '6a', '5a', '4a', '3a', '2a', '1a',
    '9b', '8b', '7b', '6b', '5b', '4b', '3b', '2b', '1b',
    '9c', '8c', '7c', '6c', '5c', '4c', '3c', '2c', '1c',
    '9d', '8d', '7d', '6d', '5d', '4d', '3d', '2d', '1d',
    '9e', '8e', '7e', '6e', '5e', '4e', '3e', '2e', '1e',
    '9f', '8f', '7f', '6f', '5f', '4f', '3f', '2f', '1f',
    '9g', '8g', '7g', '6g', '5g', '4g', '3g', '2g', '1g',
    '9h', '8h', '7h', '6h', '5h', '4h', '3h', '2h', '1h',
    '9i', '8i', '7i', '6i', '5i', '4i', '3i', '2i', '1i',
]

class Move(object):
    '''
    Represents a move from a square to a square and possibly the promotion piece
    type.
    Null moves are supported.
    '''

    def __init__(self, from_square, to_square, promotion=False, drop_piece_type=None):
        # if from_square is None, it's a drop and
        self.from_square = from_square
        self.to_square = to_square
        if from_square is None and to_square is not None:
            if drop_piece_type is None:
                raise ValueError('Drop piece type must be set.')
            if promotion:
                raise ValueError('Cannot set promoted piece.')
            self.promotion = False
            self.drop_piece_type = drop_piece_type
        else:
            self.promotion = promotion
            if drop_piece_type:
                raise ValueError('Drop piece type must not be set.')
            self.drop_piece_type = None

    def usi(self):
        '''
        Gets an USI string for the move.
        For example a move from 7A to 8A would be `7a8a` or `7a8a+` if it is
        a promotion.
        '''
        if self:
            if self.drop_piece_type:
                return '{0}*{1}'.format(PIECE_SYMBOLS[self.drop_piece_type].upper(), SQUARE_NAMES[self.to_square])
            else:
                return SQUARE_NAMES[self.from_square] + SQUARE_NAMES[self.to_square] + \
                       ('+' if self.promotion else '')
        else:
            return '0000'

    def __bool__(self):
        return not bool(self.from_square is None and self.to_square is None)

    def __nonzero__(self):
        return (self.from_square or self.to_square) is not None

    def __eq__(self, other):
        try:
            return self.from_square == other.from_square and self.to_square == other.to_square and \
                   self.promotion == other.promotion and self.drop_piece_type == other.drop_piece_type
        except AttributeError:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __repr__(self):
        return "Move.from_usi('{0}')".format(self.usi())

    def __str__(self):
        return self.usi()

    def __hash__(self):
        # 7 bit is enough to represent 81 patterns
        if self.drop_piece_type is None:
            return self.to_square | self.from_square << 7 | self.promotion << 14
        else:
            # drop piece
            return self.to_square | (81 + self.drop_piece_type) << 7

    @classmethod
    def from_usi(cls, usi):
        '''
        Parses an USI string.
        Raises `ValueError` if the USI string is invalid.
        '''
        if usi == '0000':
            return cls.null()
        elif len(usi) == 4:
            if usi[1] == '*':
                piece = Piece.from_symbol(usi[0])
                return cls(None, SQUARE_NAMES.index(usi[2:4]), False, piece.piece_type)
            else:
                return cls(SQUARE_NAMES.index(usi[0:2]), SQUARE_NAMES.index(usi[2:4]))
        elif len(usi) == 5 and usi[4] == '+':
            return cls(SQUARE_NAMES.index(usi[0:2]), SQUARE_NAMES.index(usi[2:4]), True)
        else:
            raise ValueError('expected usi string to be of length 4 or 5')

    @classmethod
    def null(cls):
        '''
        Gets a null move.
        A null move just passes the turn to the other side (and possibly
        forfeits en-passant capturing). Null moves evaluate to `False` in
        boolean contexts.
        >>> bool(shogi.Move.null())
        False
        '''
        return cls(None, None, NONE)



# -*- coding: utf-8 -*-
#
# This file is part of the python-shogi library.
# Copyright (C) 2015- Tasuku SUENAGA <tasuku-s-github@titech.ac>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals

import re
import shogi

NAME_SUFFIX_RE = re.compile(r'(小\d|奨励会|アマ|さん|[四五六七八九]段|[一二三四五六]冠|前?(名人|棋聖|王位|王座|王将|棋王|九段|十段|竜王|nhk杯|新人王|女王|女流王座|女流名人|女流王将|女流王位|倉敷藤花)+)$')

NAMES_OF_PROFESSIONAL_PLAYERS = [
    # old
    '天野宗歩',
    '小野五平',
    '溝呂木光治',
    '小菅剣之助',
    '関根金次郎',
    '花田長太郎',
    '小林東伯斎',
    '阪田三吉', '坂田三吉',
    '木見金治郎',
    '井上義雄',
    '大崎熊雄',
    '宮松関三郎',
    '川井房郷',
    '石井秀吉',
    '長谷川清二郎',
    '神田辰之助',
    '土居市太郎',
    '山本樟郎',
    '小泉雅信',
    '平野信助',
    '藤内金吾',
    '飯塚勘一郎',
    '斎藤銀次郎',
    '鈴木禎一',
    '建部和歌夫',
    '上田三三',
    '畝美与吉',
    '松田辰雄',
    '大和久彪',
    '橋爪敏太郎',
    '野村慶虎',
    '村上真一',
    '金高清吉',
    '間宮純一',
    '藤川義夫',
    '京須行男',
    '松浦卓造',
    '山田道美',
    '市川伸',
    # shogi renmei
    '金易二郎',
    '木村義雄',
    '金子金五郎',
    '渡辺東一',
    '萩原淳',
    '中井捨吉',
    '大野源一',
    '志沢春吉',
    '市川一郎',
    '坂口允彦',
    '塚田正夫',
    '梶一郎',
    '角田三男',
    '加藤治郎',
    '松下力',
    '奥野基芳',
    '小堀清一',
    '升田幸三',
    '高島一岐代',
    '荒巻三之',
    '永沢勝雄',
    '加藤恵三',
    '岡崎史明',
    '北楯修哉',
    '松田茂役',
    '大山康晴',
    '山本武雄',
    '山中和正',
    '板谷四郎',
    '本間爽悦',
    '高柳敏夫',
    '廣津久雄',
    '吉田六彦',
    '富沢幹雄',
    '原田泰夫',
    '星田啓三',
    '南口繁一',
    '北村秀治郎',
    '花村元司',
    '山川次彦',
    '佐瀬勇次',
    '加藤博二',
    '丸田祐三',
    '灘蓮照',
    '平野広吉',
    '五十嵐豊一',
    '清野静男',
    '佐藤豊',
    '木川貴一',
    '下平幸男',
    '二見敬三',
    '西本馨',
    '神田鎮雄',
    '熊谷達人',
    '増田敏二',
    '浅沼一',
    '二上達也',
    '橋本三治',
    '津村常吉',
    '北村昌男',
    '関根茂',
    '大友昇',
    '佐藤庄平',
    '加藤一二三',
    '宮坂幸雄',
    '有吉道夫',
    '長谷部久雄',
    '芹沢博文',
    '関屋喜代作',
    '賀集正三',
    '大村和久',
    '大原英二',
    '剱持松二',
    '佐藤大五郎',
    '吉田利勝',
    '北村文男',
    '内藤國雄',
    '伊達康夫',
    '佐伯昌優',
    '木村嘉孝',
    '山口千嶺',
    '木村義徳',
    '高島弘光',
    '板谷進',
    '米長邦雄',
    '大内延介',
    '西村一義',
    '木下晃',
    '山口英夫',
    '桜井昇',
    '田辺一郎',
    '中原誠',
    '桐山清澄',
    '高田丈資',
    '河口俊彦',
    '勝浦修',
    '石田和雄',
    '若松政和',
    '森安秀光',
    '森雞二', '森けい二',
    '滝誠一郎',
    '池田修一',
    '野本虎次',
    '田中魁秀',
    '坪内利幸',
    '佐藤義則',
    '安恵照剛',
    '森安正幸',
    '田丸昇',
    '宮田利男',
    '真部一男',
    '小阪昇',
    '淡路仁茂',
    '青野照市',
    '椎橋金司',
    '前田祐司',
    '飯野健二',
    '伊藤果',
    '菊地常夫',
    '桐谷広人',
    '沼春雄',
    '有野芳人',
    '小林健二',
    '土佐浩司',
    '酒井順吉',
    '森信雄',
    '田中寅彦',
    '東和男',
    '中田章道',
    '大島映二',
    '谷川浩司',
    '松浦隆一',
    '青木清',
    '小野修一',
    '福崎文吾',
    '鈴木輝彦',
    '武者野勝巳',
    '脇謙二',
    '永作芳也',
    '瀬戸博晴',
    '児玉孝一',
    '高橋道雄',
    '中村修',
    '泉正樹',
    '依田有司',
    '島朗',
    '南芳一',
    '塚田泰明',
    '神谷広志',
    '植山悦行',
    '西川慶二',
    '武市三郎',
    '室岡克彦',
    '堀口弘治',
    '大野八一雄',
    '加瀬純一',
    '井上慶太',
    '有森浩三',
    '飯田弘之',
    '神吉宏充',
    '森下卓',
    '浦野真彦',
    '小野敦生',
    '日浦市郎',
    '達正光',
    '伊藤博文',
    '小林宏',
    '富岡英作',
    '関浩',
    '本間博',
    '阿部隆',
    '所司和晴',
    '中田宏樹',
    '安西勝一',
    '羽生善治',
    '中田功',
    '石川陽生',
    '長沼洋',
    '神崎健二',
    '村山聖',
    '櫛田陽一',
    '佐藤康光',
    '森内俊之',
    '中川大輔',
    '先崎学',
    '野田敬三',
    '木下浩一',
    '小倉久史',
    '屋敷伸之',
    '藤原直哉',
    '高田尚平',
    '畠山鎮',
    '畠山成幸',
    '丸山忠久',
    '郷田真隆',
    '佐藤秀司',
    '杉本昌隆',
    '藤井猛',
    '平藤眞吾',
    '豊川孝弘',
    '深浦康市',
    '真田圭一',
    '飯塚祐紀',
    '三浦弘行',
    '伊藤能',
    '川上猛',
    '久保利明',
    '行方尚史',
    '岡崎洋',
    '窪田義行',
    '北浜健介',
    '矢倉規広',
    '鈴木大介',
    '北島忠雄',
    '勝又清和',
    '松本佳介',
    '田村康介',
    '堀口一史座',
    '中座真',
    '近藤正和',
    '野月浩貴',
    '木村一基',
    '小林裕士',
    '佐藤紳哉',
    '増田裕司',
    '高野秀行',
    '山崎隆之',
    '伊奈祐介',
    '山本真也',
    '中尾敏之',
    '松尾歩',
    '金沢孝史',
    '阿久津主税',
    '安用寺孝功',
    '渡辺明',
    '飯島栄治',
    '千葉幸生',
    '上野裕和',
    '橋本崇載',
    '佐々木慎',
    '宮田敦史',
    '村田智弘',
    '大平武洋',
    '熊坂学',
    '藤倉勇樹',
    '横山泰明',
    '島本亮',
    '西尾明',
    '村山慈明',
    '佐藤和俊',
    '片上大輔',
    '中村亮介',
    '村中秀史',
    '阪口悟',
    '広瀬章人',
    '長岡裕也',
    '高崎一生',
    '遠山雄亮',
    '瀬川晶司',
    '糸谷哲郎',
    '中村太地',
    '戸辺誠',
    '佐藤天彦',
    '豊島将之',
    '金井恒太',
    '伊藤真吾',
    '村田顕弘',
    '及川拓馬',
    '稲葉陽',
    '田中悠一',
    '佐藤慎一',
    '西川和宏',
    '吉田正和',
    '澤田真吾',
    '大石直嗣',
    '永瀬拓矢',
    '阿部健治郎',
    '菅井竜也',
    '牧野光則',
    '佐々木勇気',
    '船江恒平',
    '門倉啓太',
    '阿部光瑠',
    '高見泰地',
    '藤森哲也',
    '斎藤慎太郎',
    '八代弥',
    '上村亘',
    '石田直裕',
    '渡辺大夢',
    '千田翔太',
    '竹内雄悟',
    '石井健太郎',
    '三枚堂達也',
    '星野良生',
    '宮本広志',
    '増田康宏',
    '黒沢怜生',
    '今泉健司',
    '青嶋未来',
    '梶浦宏孝',
    '高野智史',
    '近藤誠也',
    '都成竜馬',
    '井出隼平',
    '佐々木大地',
    '藤井聡太',
    '大橋貴洸',
    '西田拓也',
    '杉本和陽',
]

NAMES_OF_LADIES_PROFESSIONAL_PLAYERS = [
    '蛸島彰子',
    '関根紀代子',
    '多田佳子',
    '山下カズ子',
    '寺下紀子',
    '村山幸子',
    '杉崎里子',
    '福崎睦美',
    '佐藤寿子',
    '谷川治恵',
    '森安多恵子',
    '宇治正子',
    '藤森奈津子',
    '長沢千和子',
    '神田真由美',
    '林葉直子',
    '中井広恵',
    '山田久美',
    '清水市代',
    '高群佐知子',
    '斎田晴子',
    '石高澄恵',
    '鹿野圭生',
    '植村真理',
    '船戸陽子',
    '大庭美樹',
    '真田彩子',
    '高橋和',
    '林まゆみ',
    '本田小百合',
    '久津知子',
    '矢内理絵子',
    '石橋幸緒',
    '中倉彰',
    '千葉涼子',
    '伊藤明日香',
    '竹部さゆり',
    '中倉宏美',
    '島井咲緒里',
    '早水千紗',
    '甲斐智美',
    '藤田麻衣子',
    '安食総子',
    '大庭美夏',
    '藤田綾',
    '上川香織',
    '野田澤彩乃',
    '山田朱未',
    '北尾まどか',
    '上田初美',
    '坂東香菜子',
    '村田智穂',
    '鈴木環那',
    '中村真梨花',
    '貞升南',
    '岩根忍',
    '里見香奈',
    '井道千尋',
    '室田伊緒',
    '伊奈川愛菓',
    '熊倉紫野',
    '中村桃子',
    '山口恵梨子',
    '香川愛生',
    '渡辺弥生',
    '室谷由紀',
    '長谷川優貴',
    '竹俣紅',
    '北村桂香',
    '相川春香',
    '飯野愛',
    '渡部愛',
    '山根ことみ',
    '和田あき',
    '塚田恵梨花',
    '伊藤沙恵',
    '中澤沙耶',
    '高浜愛子',
    '山口絵美菜',
    '里見咲紀',
    '石本さくら',
    '頼本奈菜',
    'カロリーナ・ステチェンスカ',
    '堀彩乃',
]

class Name:
    @staticmethod
    def normalize(name):
        if name is None:
            return None
        name = name.translate({
            ord(' '): None     , # space
            ord('　'): None, # full-width space
            ord('・'): None, # full-width center dot
        })
        name = NAME_SUFFIX_RE.sub('', name)

        return name

    @staticmethod
    def is_professional(name):
        normalized = Name.normalize(name)
        return normalized in NAMES_OF_PROFESSIONAL_PLAYERS

    @staticmethod
    def is_ladies_professional(name):
        normalized = Name.normalize(name)
        return normalized in NAMES_OF_LADIES_PROFESSIONAL_PLAYERS



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

from .Consts import *

PIECE_SYMBOLS = ['',   'p',  'l',  'n',  's', 'g',  'b',  'r', 'k',
                      '+p', '+l', '+n', '+s',      '+b', '+r']
PIECE_JAPANESE_SYMBOLS = [
    '',
    '歩', '香', '桂', '銀', '金', '角', '飛',
    '玉', 'と', '杏', '圭', '全', '馬', '龍'
]

class Piece(object):
    def __init__(self, piece_type, color):
        if piece_type is None:
            raise ValueError('Piece type must be set')
        if color is None:
            raise ValueError('Color must be set')
        self.piece_type = piece_type
        self.color = color

    def symbol(self):
        '''
        Gets the symbol `p`, `l`, `n`, etc.
        '''
        if self.color == BLACK:
            return PIECE_SYMBOLS[self.piece_type].upper()
        else:
            return PIECE_SYMBOLS[self.piece_type]

    def japanese_symbol(self):
        # no direction
        return PIECE_JAPANESE_SYMBOLS[self.piece_type]

    def japanese_symbol_with_direction(self):
        if self.color == BLACK:
            prefix = ' '
        else:
            prefix = 'v'
        return prefix + PIECE_JAPANESE_SYMBOLS[self.piece_type]

    def is_promoted(self):
        return self.piece_type >= PROM_PAWN

    def __hash__(self):
        return self.piece_type * (self.color + 1)

    def __repr__(self):
        return "Piece.from_symbol('{0}')".format(self.symbol())

    def __str__(self):
        return self.symbol()

    def __eq__(self, other):
        try:
            return self.piece_type == other.piece_type and self.color == other.color
        except AttributeError:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    @classmethod
    def from_symbol(cls, symbol):
        '''
        Creates a piece instance from a piece symbol.
        Raises `ValueError` if the symbol is invalid.
        '''
        if symbol.lower() == symbol:
            return cls(PIECE_SYMBOLS.index(symbol), WHITE)
        else:
            return cls(PIECE_SYMBOLS.index(symbol.lower()), BLACK)
